"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = exports.analyze = void 0;
const parser_1 = require("@babel/parser");
const traverse_1 = __importDefault(require("@babel/traverse"));
const fs_1 = __importDefault(require("fs"));
const globby_1 = __importDefault(require("globby"));
const path_1 = __importDefault(require("path"));
const sleep_1 = require("./sleep");
/**
 * Return a Promise with the JSX usage analysis of a project
 *
 * This function can easily return several megabytes of data. Parsing every
 * single JS/TS file in a project can take a while. Please be patient. The
 * `onFile` callbacks are async. Parsing each file is not an async action and
 * may block for a couple seconds if the file is large. You can insert a "sleep"
 * command into the `onFile` callback if you would like to return control
 * briefly.
 */
function analyze({ babelPlugins = [], components, directory = process.cwd(), files = ["**/*.{js,jsx,tsx}"], gitignore = true, ignore = [], onFile, prop = "", relativePaths = false, }) {
    return __awaiter(this, void 0, void 0, function* () {
        function processFilename(filename) {
            if (relativePaths) {
                return filename;
            }
            return path_1.default.resolve(directory, filename);
        }
        const searchProp = prop;
        const timeStart = Date.now();
        const filenames = yield globby_1.default(files || "**/*.{js,jsx,tsx}", {
            absolute: !relativePaths,
            onlyFiles: true,
            gitignore,
            ignore,
            cwd: directory,
        });
        const reporter = new Reporter();
        for (const filename of filenames) {
            if (onFile) {
                yield onFile(processFilename(filename));
            }
            else {
                yield sleep_1.sleep();
            }
            try {
                const code = fs_1.default.readFileSync(path_1.default.resolve(directory, filename), "utf8");
                parse(code, {
                    babelPlugins,
                    typescript: filename.endsWith(".tsx") || filename.endsWith(".ts"),
                    onlyComponents: components,
                    onComponent: ({ componentName, node }) => {
                        reporter.addComponent(componentName);
                        const props = node.openingElement.attributes.map((propNode) => {
                            if (!propNode.loc) {
                                throw new Error(`JSXElement propNode.loc is missing`);
                            }
                            return {
                                componentName,
                                propName: createProp(propNode),
                                propCode: code.slice(propNode.start || 0, propNode.end || -1),
                                startLoc: propNode.loc.start,
                                endLoc: propNode.loc.end,
                                propValue: propNode.type === "JSXAttribute"
                                    ? formatPropValue(propNode.value)
                                    : formatPropValue(null),
                            };
                        });
                        if (searchProp.startsWith("!")) {
                            const wantProp = searchProp.slice(1);
                            if (props.every((p) => p.propName !== wantProp)) {
                                if (!node.loc) {
                                    throw new Error(`JSXElement propNode.loc is missing`);
                                }
                                reporter.addProp(componentName, searchProp, {
                                    propCode: code.slice(node.start || 0, node.end || -1),
                                    startLoc: node.loc.start,
                                    endLoc: node.loc.end,
                                    prettyCode: formatPrettyCode(code, node.loc.start.line, node.loc.end.line),
                                    filename: processFilename(filename),
                                });
                            }
                        }
                        else {
                            for (const prop of props) {
                                let wantPropKey = searchProp;
                                let match = (_value) => true;
                                if (searchProp.includes("!=")) {
                                    const index = searchProp.indexOf("!=");
                                    const key = searchProp.slice(0, index);
                                    const val = searchProp.slice(index + 2);
                                    wantPropKey = key;
                                    match = (value) => value !== val;
                                }
                                else if (searchProp.includes("=")) {
                                    const index = searchProp.indexOf("=");
                                    const key = searchProp.slice(0, index);
                                    const val = searchProp.slice(index + 1);
                                    wantPropKey = key;
                                    match = (value) => value === val;
                                }
                                if (wantPropKey && prop.propName !== wantPropKey) {
                                    return;
                                }
                                if (!match(prop.propValue)) {
                                    return;
                                }
                                reporter.addProp(componentName, prop.propName, {
                                    propCode: prop.propCode,
                                    startLoc: prop.startLoc,
                                    endLoc: prop.endLoc,
                                    prettyCode: formatPrettyCode(code, prop.startLoc.line, prop.endLoc.line),
                                    filename: processFilename(filename),
                                });
                            }
                        }
                    },
                });
            }
            catch (error) {
                if (error instanceof SyntaxError) {
                    reporter.addParseError(processFilename(filename), error);
                }
                else {
                    throw error;
                }
            }
        }
        const elapsedTime = (Date.now() - timeStart) / 1000;
        return {
            filenames,
            componentTotal: reporter.getComponentTotal(),
            componentUsageTotal: reporter.getComponentUsageTotal(),
            componentUsage: reporter.components,
            propUsage: reporter.props,
            lineUsage: reporter.lines,
            errors: reporter.errors,
            suggestedPlugins: reporter.suggestedPlugins,
            elapsedTime: elapsedTime,
            directory,
        };
    });
}
exports.analyze = analyze;
////////////////////////////////////////////////////////////////////////////////
//
// Implementation details below
//
////////////////////////////////////////////////////////////////////////////////
const linesCache = new Map();
function getLines(code) {
    let lines = linesCache.get(code);
    if (lines) {
        return lines;
    }
    lines = code.split(/\r?\n/);
    linesCache.set(code, lines);
    return lines;
}
const EXPRESSION = Symbol("formatPropValue.EXPRESSION");
function formatPropValue(value) {
    if (value === null) {
        return "true";
    }
    if (!value) {
        return EXPRESSION;
    }
    switch (value.type) {
        // TODO: Should we interpret anything else here?
        case "StringLiteral":
            return value.value;
        case "JSXExpressionContainer":
            return formatPropValue(value.expression);
        case "NumericLiteral":
        case "BooleanLiteral":
            return String(value.value);
        default:
            return EXPRESSION;
    }
}
function getAttributeName(attributeNode) {
    switch (attributeNode.type) {
        case "JSXIdentifier":
            return attributeNode.name;
        case "JSXNamespacedName":
            return attributeNode.name.name;
        case "JSXAttribute":
            return getAttributeName(attributeNode.name);
        case "JSXSpreadAttribute":
            return "{...}";
        default:
            throw new Error(`unexpected node type: ${attributeNode.type}`);
    }
}
function createProp(attributeNode) {
    return getAttributeName(attributeNode);
}
function getDottedName(nameNode) {
    switch (nameNode.type) {
        case "JSXMemberExpression":
            return [nameNode.object, nameNode.property].map(getDottedName).join(".");
        case "JSXIdentifier":
            return nameNode.name;
        default:
            throw new Error(`unexpected node type: ${nameNode.type}`);
    }
}
function createComponent(componentNode) {
    return getDottedName(componentNode.openingElement.name);
}
function parse(code, options = {}) {
    const { typescript = false, babelPlugins = [], onlyComponents = [], onComponent = () => { }, } = options;
    function doReportComponent(component) {
        if (onlyComponents.length === 0) {
            return true;
        }
        return onlyComponents.indexOf(component) !== -1;
    }
    const ast = parser_1.parse(code, {
        sourceType: "unambiguous",
        allowReturnOutsideFunction: true,
        plugins: [
            typescript ? "typescript" : "flow",
            "jsx",
            "dynamicImport",
            "classProperties",
            "objectRestSpread",
            ...babelPlugins,
        ],
    });
    traverse_1.default(ast, {
        JSXElement(path) {
            const node = path.node;
            const componentName = createComponent(node);
            if (doReportComponent(componentName)) {
                onComponent({ componentName, node });
            }
        },
    });
}
exports.parse = parse;
function formatPrettyCode(code, startLine, endLine) {
    const output = [];
    const lines = getLines(code);
    // Line numbers should be padded to at least 4 digits for consistency and
    // readability, but let's also let them grow if we have super long files :|
    const maxDigits = Math.max(String(startLine).length, String(endLine).length, 4);
    for (let lineno = startLine; lineno <= endLine; lineno++) {
        output.push(String(lineno).padStart(maxDigits) + " | " + lines[lineno - 1]);
    }
    return output.join("\n");
}
class Reporter {
    constructor() {
        this.components = {};
        this.props = {};
        this.lines = {};
        this.suggestedPlugins = [];
        this.errors = {};
    }
    // TODO: Does Babel expose the actual error type for us?
    addParseError(filename, error) {
        this.errors[filename] = {
            message: error.message,
            pos: error.pos,
            loc: error.loc,
            missingPlugin: error.missingPlugin,
        };
        for (const plugin of error.missingPlugin || []) {
            if (!this.suggestedPlugins.includes(plugin)) {
                this.suggestedPlugins.push(plugin);
            }
        }
    }
    addComponent(componentName) {
        this.components[componentName] = (this.components[componentName] || 0) + 1;
    }
    _incrementProp(componentName, propName) {
        if (!this.props[componentName]) {
            this.props[componentName] = {};
        }
        const count = this.props[componentName][propName] || 0;
        this.props[componentName][propName] = count + 1;
    }
    _ensureLines(componentName, propName) {
        if (!this.lines[componentName]) {
            this.lines[componentName] = {};
        }
        if (!this.lines[componentName][propName]) {
            this.lines[componentName][propName] = [];
        }
    }
    addProp(componentName, propName, line) {
        this._incrementProp(componentName, propName);
        this._ensureLines(componentName, propName);
        this.lines[componentName][propName].push(line);
    }
    getComponentTotal() {
        return Object.keys(this.components).length;
    }
    getComponentUsageTotal() {
        return Object.values(this.components).reduce((a, b) => a + b, 0);
    }
}
