"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.report = exports.files = exports.ignore = exports.gitignore = exports.directory = exports.babelPlugins = exports.showProgress = exports.components = exports.prop = void 0;
const chalk_1 = __importDefault(require("chalk"));
const commander_1 = __importDefault(require("commander"));
const cosmiconfig_1 = require("cosmiconfig");
const path_1 = __importDefault(require("path"));
const package_json_1 = require("../package.json");
const printer_1 = require("./printer");
commander_1.default.name("jsx-info");
commander_1.default.version(package_json_1.version, "-v, --version");
commander_1.default
    .description("Displays a report of JSX component and prop usage")
    .option("--no-progress", "don't show progress messages")
    .option("--components <components...>", "which components to scan")
    .option("--color", "force enable color")
    .option("--no-color", "force disable color")
    .option("--no-config", "disable config file")
    .option("--no-gitignore", "disable reading .gitignore files")
    .option("--babel-plugins <plugins...>", "set Babel plugins")
    .option("--directory <directory>", "directory to use as the base for finding files")
    .option("--ignore <patterns...>", "glob patterns used to ignore input files")
    .option("--files <patterns...>", "glob patterns used to find input files")
    .option("--report <usage|props|lines>", "which report to show")
    .option("--prop <prop>", "which prop to search for when running a lines report (e.g. `id` or `variant=primary` or `!className` or `type!=text`)");
commander_1.default.on("--help", () => {
    printer_1.print(chalk_1.default `
Examples:
  Use jsx-info interactively:
      {bold.cyan npx jsx-info}

  Display info only for <div> and <Tab.Container>
      {bold.cyan npx jsx-info --components} {magenta div Tab.Container}

  See lines where className prop was used on <div> component
      {bold.cyan npx jsx-info --report} {magenta lines} {bold.cyan --prop} {magenta className} {bold.cyan --components} {magenta div}

  See lines where \`id\` prop was used on any component
      {bold.cyan npx jsx-info --report} {magenta lines} {bold.cyan --prop} {magenta id} {bold.cyan --components} {magenta "*"}

  See lines where the kind prop was used with value "primary" on Button component
      {bold.cyan npx jsx-info --report} {magenta lines} {bold.cyan --prop} {magenta kind=primary} {bold.cyan --components} {magenta Button}

  See lines where the \`kind\` prop was used with any value besides "primary" on the Button component
      {bold.cyan npx jsx-info --report} {magenta lines} {bold.cyan --prop} {magenta 'kind!=primary'} {bold.cyan --components} {magenta Button}

  See lines where the \`kind\` prop wasn't used on the Button component
      {bold.cyan npx jsx-info --report} {magenta lines} {bold.cyan --prop} {magenta '!kind'} {bold.cyan --components} {magenta Button}

  Ignore folders based on glob paths
      {bold.cyan npx jsx-info --ignore} {magenta "**/__test__"} {magenta packages/legacy}

  Enable Babel plugins
      {bold.cyan npx jsx-info --babel-plugins} {magenta decorators-legacy} {magenta pipelineOperator}

  Example .jsx-info.json config file
      ${"{"}
        {cyan "babelPlugins"}: [{magenta "decorators-legacy"}, {magenta "pipelineOperator"}],
        {cyan "directory"}: {magenta "src"},
        {cyan "ignore"}: [{magenta "**/__test__"}, {magenta "legacy/**"}],
        {cyan "files"}: [{magenta "**/*.${"{"}js,jsx,tsx${"}"}"}]
      ${"}"}
`);
});
commander_1.default.parse(process.argv);
printer_1.print(chalk_1.default.magenta.bold(`jsx-info ${package_json_1.version} by @wavebeem`));
function getConfig() {
    if (!commander_1.default.config) {
        return {};
    }
    try {
        const explorer = cosmiconfig_1.cosmiconfigSync("jsx-info", {
            searchPlaces: ["package.json", ".jsx-info.json"],
        });
        const result = explorer.search();
        if (result) {
            printer_1.print(`\nLoaded configuration from ${result.filepath}\n`);
            if (result.config.directory) {
                result.config.directory = path_1.default.resolve(result.filepath, "..", result.config.directory);
            }
            return result.config;
        }
    }
    catch (err) {
        if (process.env.DEBUG) {
            // eslint-disable-next-line no-console
            console.log(err.stack);
        }
        printer_1.printError("failed to parse config file");
    }
    return {};
}
const config = getConfig();
function concat(a = [], b = []) {
    return [...a, ...b];
}
exports.prop = commander_1.default.prop;
exports.components = commander_1.default.components;
exports.showProgress = commander_1.default.progress;
exports.babelPlugins = concat(config.babelPlugins, commander_1.default.babelPlugins);
exports.directory = commander_1.default.directory || config.directory;
exports.gitignore = commander_1.default.gitignore;
exports.ignore = concat(commander_1.default.ignore, config.ignore);
exports.files = concat(commander_1.default.files, config.files);
exports.report = commander_1.default.report;
